# -*- perl -*-

# Copyright (c) 2002 by Jeff Weisberg
# Author: Jeff Weisberg <argus @ tcp4me.com>
# Date: 2002-Dec-30 23:29 (EST)
# Function: configurable objects
#
# $Id: Configable.pm,v 1.3 2003/09/14 21:20:20 jaw Exp $

package Configable;

$doc = {
    package => __PACKAGE__,
    file    => __FILE__,
    isa     => [ ],
    methods => {},
    versn   => '3.2',
    fields  =>  {
	name  => {},
	type  => {},
	quotp => {},
	subtypes => {},
    },
};

# initialize from config data according to method described in $doc
# hmmm, perhaps $doc is a poor choice of names....
sub init_from_config {
    my $me   = shift;
    my $doc  = shift;
    my $base = shift;

    foreach my $k (keys %{$doc->{fields}}){
	# ignore fields that are not marked configurable
	next unless has_attr($k, $doc, 'config');
	# next if has_attr($k, $doc, 'top');
	$me->init_field_from_config($doc, $base, $k);
    }
}

sub init_field_from_config {
    my $me   = shift;
    my $doc  = shift;
    my $base = shift;
    my $fld  = shift;
    my( $v, $kk );

    $kk = $fld;
    $kk =~ s/^($base)::// if $base;
    return if( $kk =~ /::/ );
    return if( $base && $fld !~ /::/ );

    # for detecting config file typos
    $me->{confck}{$kk} = 1;
    
    # does this field inherit from parent if not set?
    if( has_attr($fld, $doc, 'inherit') ){
	$v = $me->inherit( $kk );
    }elsif( has_attr($fld, $doc, 'acl') ){
	$v = $me->build_acl($kk);
    }else{
	$v = $me->{config}{$kk};
    }

    # if not set, is there a default
    $v = $doc->{fields}{$fld}{default} unless defined($v);
    
    if( has_attr($fld, $doc, 'bool') ){
	$v = ::ckbool($v) if defined($v);
    }

    # install value
    if( $base ){
	$me->{$base}{$kk} = $v if defined($v);
    }else{
	$me->{$kk} = $v if defined($v);
    }

}

# And he said unto him, I am the LORD that brought thee out of
# Ur of the Chaldees, to give thee this land to inherit it.
#     -- genesis 15:7
sub inherit {
    my $me = shift;
    my $field = shift;

    $me->{confck}{$field} = 1;
    return $me->{config}{$field} if defined $me->{config}{$field};
    # special case Host "hostname" {...}
    return $me->{name} if( $me->{type} eq 'Host' && $field eq 'hostname' );
    return $me->{parents}[0]->inherit($field) if $me->{parents}[0];
    return $Conf::Config{$field} if defined $Conf::Config{$field};
    undef;
}

# build cumulative acl
sub build_acl {
    my $me = shift;
    my $field = shift;
    my( $pv, $mv, $v, %g );

    # parent values, then mine
    if( $me->{parents}[0] ){
	$pv = $me->{parents}[0]->build_acl($field) || '';
	# yes, this could be done smarter, ...
    }else{
	$pv = $Conf::Config{$field} || '';
    }
    $mv = $me->{config}{$field} || '';
    $v = "$pv $mv";

    # process directives, etc
    foreach my $g (split /\s+/, $v){
	if( $g eq '-ALL' ){
	    %g = ();
	    next;
	}
	if( $g =~ /^-(.*)/ ){
	    delete $g{$1};
	    next;
	}
	$g{$g} = 1;
    }

    $v = join ' ', keys %g;

    $v || undef;
}

# has_attr gets called for every field for every object (or about 500 times per object)
# at a 1000 objects, this adds up to some significant time
# memoize...
my %hasattrcache = ();

sub has_attr {
    my $field = shift;
    my $doc   = shift;
    my $attr  = shift;
    my( $v );
    
    $v = $hasattrcache{ $doc->{package} }{$field}{$attr};
    return $v if defined $v;
    
    $v = grep {/$attr/} @{$doc->{fields}{$field}{attrs}};
    $hasattrcache{ $doc->{package} }{$field}{$attr} = $v;
    $v;
}

sub clearcache {
    %hasattrcache = ();
}

# Error has no end.
#   -- Robert Browning, Paracelsus. Part iii.
sub check_typos {
    my $me = shift;
    my $cf = shift;

    return unless $cf;
    foreach my $k (keys %{$me->{config}}){
	next if $me->{confck}{$k};
	$cf->warning( "unused parameter '$k' - typo?" );
    }
}

sub gen_conf {
    my $me = shift;
    my( $r, $d );

    $d = Doc::objdocs($me);
    $r = $me->{type}  . ' ';
    $r .= $d->{conf}{quotp} ? "\"$me->{name}\"" : $me->{name};
    if( $d->{conf}{bodyp} || keys %{$me->{config}} ){
	$r .= " {\n";
	foreach my $k (sort keys %{$me->{config}}){
	    my $v = $me->{config}{$k};
	    $v =~ s/\#/\\\#/g;
	    $v =~ s/\n/\\n\\\n/g;
	    $r .= "\t$k:\t$v";
	    $r .= "\t# unused parameter - typo?"
		unless $me->{confck}{$k} || $d->{conf}{notypos};
	    $r .= "\n";
	}
	if(exists $me->{children} || $me->{cronjobs}){
	    my $rc;
	    foreach my $c (@{$me->{cronjobs}}, @{$me->{children}}){
		$rc .= $c->gen_conf();
	    }
	    if( $rc ){
		$rc =~ s/^/\t/gm;
		$r .= "$rc";
	    }
	}
	$r .= "}";
    }
    
    $r .= "\n";
    $r;
}

# XXX not yet used
sub readconfig2 {
    my $class = shift;
    my $cf  = shift;
    my $mom = shift;
    my $me  = $class->new;
    my( $d, $line, $type, $name, $x, $nomoredata, @kinder );

    print STDERR "rc2 test\n";
    
    $d = Doc::objdocs($me);
    $line = $cf->nextline();
    ($type, $name) = $line =~ /(\S+)\s+\"(.*)\"/i;
    $me->{type} = "\u\L$type";
    $me->{name} = $name;
    $me->{parents} = [ $mom ] if $mom;
    $me->{definedinfile} = $cf->{file};
    $me->{definedonline} = $cf->{line};
    $me->{definedattime} = $^T;

    my $permitted = $type =~ /$d->{conf}{subtypes}/;
    
    while( defined($_ = $cf->nextline()) ){
	print STDERR "readconfig2: $_\n" if $::opt_d;
	if( /^\s*\}/ ){
	    $me->{children} = [ @kinder ];
	    $me->config($cf);
	    return $me;
	}

	elsif( /^(Group|Host)\s/i && $permitted ){
	    $nomoredata = 1;
	    $cf->ungetline($_);
	    $x = Group->readconfig2($cf, $me);
	    push @kinder, $x if $x;
	}

	elsif( /^Alias\s/i && $permitted ){
	    $nomoredata = 1;
	    $cf->ungetline($_);
	    $x = Alias->readconfig2($cf, $me);
	    push @kinder, $x if $x;
	}
	elsif( /^Service/i && $permitted ){	# no, not Service\s, we permit a ':' for symmetry
	    $nomoredata = 1;
	    $cf->ungetline($_);
	    $x = Service->readconfig2($cf, $me);
	    push @kinder, $x if $x;
	}
	elsif( /^Method\s/i && $permitted ){
	    $me->ungetline($_);
	    NotMe->readconfig2($cf, $me);
	}
	elsif( /^cron\s/i && $permitted ){
	    $cf->ungetline($_);
	    $x = UserCron->readconfig2($cf, $me);
	    push @{ $me->{cronjobs} }, $x if $x;
	}

	elsif( /:/ ){
	    my($k, $v) = split /:\s*/, $_, 2;
	    # data must be before children
	    # disallow
	    # Group "foo" {
	    #    x1: value
	    #    Service TCP/FOO
	    #    x2: value          <- error
	    # }
	    if( $nomoredata ){
		$cf->warning( "additional data not permitted here (ignored)" );
		next;
	    }
	    $cf->warning( "redefinition of parameter '$k'" )
		if defined $me->{config}{$k};
	    $me->{config}{$k} = $v;
	}

	else{
	    return $cf->error( "invalid entry in config file: '$_'" );
	}
    }

    $me;
}


1;

